import ssl, time
from pathlib import Path

import random
import string

import paho.mqtt.client as mqtt
from terminaltables import SingleTable

from mqtt_client.subscribe_callbacks import (
    default_subscribe_callback,
    subscribe_callback_raw,
    subscribe_callback_command,
)

CERT_DEFAULT_PATH = 'mqtt_broker_cert.pem'
CONNECT_MQTT_BROKER = False
TIMEOUT_DEFAULT = 5


class MqttWrapper:
    def __init__(self, host, port, topic, auth, client_id=False, transport='tcp'):
        self.host = host
        self.port = port
        self.auth = auth
        self.timeout = TIMEOUT_DEFAULT
        self.topic = topic
        self._set_transport(transport=transport)
        self.cert_path = CERT_DEFAULT_PATH

        clean_session = False
        if client_id is False:
          client_id = 'mqtt-client-' + ''.join(random.choice(string.ascii_lowercase) for i in range(6))
          clean_session = True

        self.client_id = client_id
        self.client = mqtt.Client(self.client_id, clean_session, transport=self.transport)
        self.client.on_connect = self.on_connect

    def _set_transport(self, transport):
        if 'TCP' == transport or 'tcp' == transport:
            self.transport, self.tls = 'tcp', False
        elif 'TCP-TLS' == transport or 'tcp-tls' == transport:
            self.transport, self.tls = 'tcp', True
        elif 'WS' == transport or 'ws' == transport:
            self.transport, self.tls = 'websocket', False
        elif 'WS-TLS' == transport or 'ws-tls' == transport:
            self.transport, self.tls = 'websocket', True

    def set_tls(self, cert_path=None):
        if cert_path:
            self.cert_path = cert_path
        my_file = Path(self.cert_path)
        if not my_file.is_file():
            self.cert_path = None

        self.client.tls_set(self.cert_path, certfile=None, keyfile=None,
                            cert_reqs=ssl.CERT_REQUIRED, tls_version=ssl.PROTOCOL_TLS, ciphers=None)

        # print(f'- SET TLS: {self.cert_path}')

    def connect(self):
        if 'username' in self.auth and 'password' in self.auth:
            if self.auth['username'] and self.auth['password']:
                self.client.username_pw_set(username=self.auth['username'], password=self.auth['password'])
        self.client.connect(self.host, self.port, self.timeout)

    def on_message(self, func):
        self.client.on_message = func

    def on_connect(self, mqttc, obj, flags, rc):
        if rc != 0:
            print('│ERROR│ from connect - rc: {}'.format(rc))

    def loop_start(self):
        self.client.loop_start()

    def loop_forever(self):
        try:
            self.client.loop_forever()
        except KeyboardInterrupt:
            exit('│CTRL+C│ Exit by KeyboardInterrupt')

    def publish(self, payload, qos=0, retain=False):
        try:
            message_info = self.client.publish(self.topic, payload=payload, qos=qos, retain=retain)
            message_info.wait_for_publish()
            return message_info.is_published()
        except Exception as ex:
            exit(ex)


def connect_to_broker(host, port, topic, username, password, client_id=False, transport='tcp', cert_path=None):
    mqtt_handler = MqttWrapper(
        host=host,
        port=port,
        topic=topic,
        auth={'username': username, 'password': password},
        client_id=client_id,
        transport=transport
    )

    table_data = [
        ['KEY', 'VALUE'],
        ['BROKER SETTINGS', f'{transport}://{host}:{port}'],
        ['CREDENTIALS USER/PASSWORD', f'{username if username else "-"} {password if password else "-"}'],
        ['CLIENT-ID', f'{mqtt_handler.client_id}'],
        ['TOPIC', f'{topic}']
    ]

    print(SingleTable(table_data).table)

    if mqtt_handler.tls:
        if cert_path:
            mqtt_handler.set_tls(cert_path=cert_path)
        else:
            mqtt_handler.set_tls()

    return mqtt_handler


def publish(mqtt_handler, payload, qos=0, retain=False):
    is_published = mqtt_handler.publish(payload=str(payload), qos=qos, retain=retain)
    table_data = [[f'publish to {mqtt_handler.topic}', payload, 'Published: OK' if is_published else False]]
    print(SingleTable(table_data).table)

    return is_published


def subscribe(mqtt_handler, callback, command):
    mqtt_handler.client.subscribe(mqtt_handler.topic)

    if not callback or callback == 'default':
        callback = default_subscribe_callback

    if callback == 'raw':
        callback = subscribe_callback_raw

    if callback == 'command' and command:
        callback = subscribe_callback_command(command=command)

    mqtt_handler.on_message(func=callback)

    table_data = [[f'waiting from {callback}', '...']]
    print(SingleTable(table_data).table)
    mqtt_handler.loop_forever()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Usage:
  mqtt-client (publish | subscribe) --config=<config>
  mqtt-client publish --host=<host> --topic=<topic> (--payload=<payload> | --interactive) [--client_id=<client_id>] [--username=<username>] [--password=<password>] [--transport=<transport>] [--cert_path=<cert_path>] [--qos=<qos>] [--retain=<retain>]
  mqtt-client subscribe --host=<host> --topic=<topic> [--client_id=<client_id>] [--username=<username>] [--password=<password>] [--transport=<transport>] [--cert_path=<cert_path>] [--callback=<callback>] [--command=<command>]
  mqtt-client (-h|--help)
  mqtt-client (-v|--version)

Commands:
  publish                   Publish to topic from MQTT Broker.
  subscribe                 Subscribe to topic from MQTT Broker.

Options:
  -h --help                 Show this screen.
  -v --version              Show version.
  --config=<config>         Config file.
  --host=<host>             Broker Host. (Example: example.your_broker.com:1883)
  --topic=<topic>           Topic.
  --payload=<payload>       Payload to send.
  -i --interactive          Interactive mode.
  --client_id=<client_id>   Client ID.
  --username=<username>     Username.
  --password=<password>     Password.
  --transport=<transport>   TCP, TCP-TLS, WS, WS-TLS (Default: TCP)
  --cert_path=<cert_path>   Path cert (Default: ./mqtt_broker_cert.pem)
  --qos=<qos>               Qos (Default: 0)
  --retain=<retain>         Retain (Default: false)
  --callback=<callback>     Use a custom callback for subscriber. (default, raw, command)
  --command=<command>       Command for callback type command.

"""
from docopt import docopt
import readline
import json

from terminaltables import SingleTable

from mqtt_client import mqtt_client

NAME, VERSION = 'MQTT Client', '1.6.0'
AUTHOR = 'Samuel de Ancos (2018-2022) <https://github.com/sdeancos/mqtt-client>'


def main():
    arguments = docopt(f'{NAME} {VERSION}\n{AUTHOR}\n\n{__doc__}', version=f'{NAME} {VERSION}')

    config = None
    if arguments['--config']:
        with open(arguments['--config']) as f:
            _config_content = f.read()
        config = json.loads(_config_content)
        host, port = config.get('host', 'localhost:1883').split(':')
        port = int(port)
        topic = config.get('topic')
        client_id = config.get('client_id', False)
        transport = config.get('transport', 'TCP')
        path = config.get('cert_path')
        username, password = config.get('username'), config.get('password')
        callback, command = config.get('callback'), config.get('command')
    else:
        host, port = 'localhost', 1883
        topic = None
        client_id = False
        transport = 'TCP'
        path = None
        username, password = None, None
        callback, command = None, None

    if arguments['--host']:
        try:
            host, port = arguments['--host'].split(':')
        except Exception:
            exit('│ERROR│ broker host failed. Example: example.your_broker.com:1883')
        if port:
            port = int(port)

    if '--client_id' in arguments and arguments['--client_id']:
        client_id = arguments['--client_id']

    if '--transport' in arguments and arguments['--transport']:
        transport = arguments['--transport']

    if '--cert_path' in arguments and arguments['--cert_path']:
        path = arguments['--cert_path']

    if '--username' in arguments and arguments['--username']:
        username = arguments['--username']

    if '--password' in arguments and arguments['--password']:
        password = arguments['--password']

    if '--callback' in arguments and arguments['--callback']:
        callback = arguments['--callback']

    if '--command' in arguments and arguments['--command']:
        command = arguments['--command']

    print(SingleTable([[NAME, VERSION]]).table)

    if not topic:
        topic = arguments['--topic']

    try:
        mqtt_handler = mqtt_client.connect_to_broker(
            host=host,
            port=port,
            topic=topic,
            client_id=client_id,
            username=username,
            password=password,
            transport=transport,
            cert_path=path
        )
        mqtt_handler.connect()
    except Exception as ex:
        exit(f'│ERROR│ {ex}')

    if arguments['publish']:
        if config and 'qos' in config:
            qos = config.get('qos', 0)
        else:
            qos = 0
            if '--qos' in arguments and arguments['--qos']:
                qos = int(arguments['--qos'])

        if config and 'retain' in config:
            retain = bool(config.get('retain', False))
        else:
            retain = 0
            if '--retain' in arguments and arguments['--retain']:
                retain = bool(arguments['--retain'])

        if (config and not 'interactive' in config) or not arguments['--interactive']:
            if config and 'payload' in config:
                payload = config['payload']
            elif '--payload' in arguments:
                payload = arguments['--payload']
            else:
                exit(f'│ERROR│ Not payload defined')

            is_published = mqtt_client.publish(mqtt_handler=mqtt_handler, payload=payload, qos=qos, retain=retain)
        else:
            while True:
                try:
                    payload = input('[insert payload] ? ')
                    is_published = mqtt_client.publish(mqtt_handler=mqtt_handler, payload=payload,
                                                       qos=qos, retain=retain)
                except KeyboardInterrupt:
                    exit('[CTRL+C] Exit')
                except EOFError:
                    exit('[CTRL+D] Exit')

    if arguments['subscribe']:
        mqtt_client.subscribe(mqtt_handler=mqtt_handler, callback=callback, command=command)

if __name__ == '__main__':
    main()



import subprocess


def default_subscribe_callback(mqttc, obj, msg):
    try:
        print(f'│{msg.topic}│ payload: {msg.payload}')
    except Exception as ex:
        print(f'Error: {ex}')


def subscribe_callback_raw(mqttc, obj, msg):
    try:
        print(msg.payload.decode('utf8'))
    except Exception as ex:
        print(f'Error: {ex}')


def subscribe_callback_command(command):
    command = [command]

    def _(mqttc, obj, msg):
        command.append(msg.topic)
        command.append(msg.payload.decode('utf8'))
        #print(f'[COMMAND] {command}')

        response = subprocess.run(command, stdout=subprocess.PIPE)
        if response.stdout:
            print(response.stdout)
        if response.stderr:
            print(response.stderr)
        command.pop(1)
        command.pop(1)

    return _



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='mqtt-client',
    version='1.6.0',
    description='Simple MQTT Client.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='mqtt',
    author='Samuel de Ancos',
    author_email='sdeancos@gmail.com',
    url='https://mqtt.clubpulp.com/',
    packages=find_packages(),
    include_package_data=False,
    python_requires='>=3.6',
    install_requires=[
        'docopt',
        'paho-mqtt',
        'terminaltables'
    ],
    entry_points={
        "console_scripts": [
            "mqtt-client = mqtt_client.__main__:main",
        ]
    },
)
